import numpy as np
from scipy.stats import percentileofscore


def score_ci(
    y_true,
    y_pred,
    score_fun,
    sample_weight=None,
    n_bootstraps=2000,
    confidence_level=0.95,
    seed=None,
    reject_one_class_samples=True,
):
    """
    Compute confidence interval for given score function based on labels and predictions using bootstrapping.
    :param y_true: 1D list or array of labels.
    :param y_pred: 1D list or array of predictions corresponding to elements in y_true.
    :param score_fun: Score function for which confidence interval is computed. (e.g. sklearn.metrics.accuracy_score)
    :param sample_weight: 1D list or array of sample weights to pass to score_fun, see e.g. sklearn.metrics.roc_auc_score.
    :param n_bootstraps: The number of bootstraps. (default: 2000)
    :param confidence_level: Confidence level for computing confidence interval. (default: 0.95)
    :param seed: Random seed for reproducibility. (default: None)
    :param reject_one_class_samples: Whether to reject bootstrapped samples with only one label. For scores like AUC we
    need at least one positive and one negative sample. (default: True)
    :return: Score evaluated on labels and predictions, lower confidence interval, upper confidence interval, array of
    bootstrapped scores.
    """

    assert len(y_true) == len(y_pred)

    score = score_fun(y_true, y_pred)
    _, ci_lower, ci_upper, scores = score_stat_ci(
        y_true=y_true,
        y_preds=y_pred,
        score_fun=score_fun,
        sample_weight=sample_weight,
        n_bootstraps=n_bootstraps,
        confidence_level=confidence_level,
        seed=seed,
        reject_one_class_samples=reject_one_class_samples,
    )

    return score, ci_lower, ci_upper, scores


def score_stat_ci(
    y_true,
    y_preds,
    score_fun,
    stat_fun=np.mean,
    sample_weight=None,
    n_bootstraps=2000,
    confidence_level=0.95,
    seed=None,
    reject_one_class_samples=True,
):
    """
    Compute confidence interval for given statistic of a score function based on labels and predictions using
    bootstrapping.
    :param y_true: 1D list or array of labels.
    :param y_preds: A list of lists or 2D array of predictions corresponding to elements in y_true.
    :param score_fun: Score function for which confidence interval is computed. (e.g. sklearn.metrics.accuracy_score)
    :param stat_fun: Statistic for which confidence interval is computed. (e.g. np.mean)
    :param sample_weight: 1D list or array of sample weights to pass to score_fun, see e.g. sklearn.metrics.roc_auc_score.
    :param n_bootstraps: The number of bootstraps. (default: 2000)
    :param confidence_level: Confidence level for computing confidence interval. (default: 0.95)
    :param seed: Random seed for reproducibility. (default: None)
    :param reject_one_class_samples: Whether to reject bootstrapped samples with only one label. For scores like AUC we
    need at least one positive and one negative sample. (default: True)
    :return: Mean score statistic evaluated on labels and predictions, lower confidence interval, upper confidence
    interval, array of bootstrapped scores.
    """

    y_true = np.array(y_true)
    y_preds = np.atleast_2d(y_preds)
    assert all(len(y_true) == len(y) for y in y_preds)

    np.random.seed(seed)
    scores = []
    for i in range(n_bootstraps):
        readers = np.random.randint(0, len(y_preds), len(y_preds))
        indices = np.random.randint(0, len(y_true), len(y_true))
        if reject_one_class_samples and len(np.unique(y_true[indices])) < 2:
            continue
        reader_scores = []
        for r in readers:
            if sample_weight is not None:
                reader_scores.append(score_fun(y_true[indices], y_preds[r][indices], sample_weight=sample_weight[indices]))
            else:
                reader_scores.append(score_fun(y_true[indices], y_preds[r][indices]))
        scores.append(stat_fun(reader_scores))

    mean_score = np.mean(scores)
    sorted_scores = np.array(sorted(scores))
    alpha = (1.0 - confidence_level) / 2.0
    ci_lower = sorted_scores[int(round(alpha * len(sorted_scores)))]
    ci_upper = sorted_scores[int(round((1.0 - alpha) * len(sorted_scores)))]
    return mean_score, ci_lower, ci_upper, scores


def pvalue(
    y_true,
    y_pred1,
    y_pred2,
    score_fun,
    sample_weight=None,
    n_bootstraps=2000,
    two_tailed=True,
    seed=None,
    reject_one_class_samples=True,
):
    """
    Compute p-value for hypothesis that score function for model I predictions is higher than for model II predictions
    using bootstrapping.
    :param y_true: 1D list or array of labels.
    :param y_pred1: 1D list or array of predictions for model I corresponding to elements in y_true.
    :param y_pred2: 1D list or array of predictions for model II corresponding to elements in y_true.
    :param score_fun: Score function for which confidence interval is computed. (e.g. sklearn.metrics.accuracy_score)
    :param sample_weight: 1D list or array of sample weights to pass to score_fun, see e.g. sklearn.metrics.roc_auc_score.
    :param n_bootstraps: The number of bootstraps. (default: 2000)
    :param two_tailed: Whether to use two-tailed test. (default: True)
    :param seed: Random seed for reproducibility. (default: None)
    :param reject_one_class_samples: Whether to reject bootstrapped samples with only one label. For scores like AUC we
    need at least one positive and one negative sample. (default: True)
    :return: Computed p-value, array of bootstrapped differences of scores.
    """

    assert len(y_true) == len(y_pred1)
    assert len(y_true) == len(y_pred2)

    return pvalue_stat(
        y_true=y_true,
        y_preds1=y_pred1,
        y_preds2=y_pred2,
        score_fun=score_fun,
        sample_weight=sample_weight,
        n_bootstraps=n_bootstraps,
        two_tailed=two_tailed,
        seed=seed,
        reject_one_class_samples=reject_one_class_samples,
    )


def pvalue_stat(
    y_true,
    y_preds1,
    y_preds2,
    score_fun,
    stat_fun=np.mean,
    compare_fun=np.subtract,
    sample_weight=None,
    n_bootstraps=2000,
    two_tailed=True,
    seed=None,
    reject_one_class_samples=True,
):
    """
    Compute p-value for hypothesis that given statistic of score function for model I predictions is higher than for
    model II predictions using bootstrapping.
    :param y_true: 1D list or array of labels.
    :param y_preds1: A list of lists or 2D array of predictions for model I corresponding to elements in y_true.
    :param y_preds2: A list of lists or 2D array of predictions for model II corresponding to elements in y_true.
    :param score_fun: Score function for which confidence interval is computed. (e.g. sklearn.metrics.accuracy_score)
    :param stat_fun: Statistic for which p-value is computed. (e.g. np.mean)
    :param compare_fun: Function to determine relative performance. (default: score1 - score2)
    :param sample_weight: 1D list or array of sample weights to pass to score_fun, see e.g. sklearn.metrics.roc_auc_score.
    :param n_bootstraps: The number of bootstraps. (default: 2000)
    :param two_tailed: Whether to use two-tailed test. (default: True)
    :param seed: Random seed for reproducibility. (default: None)
    :param reject_one_class_samples: Whether to reject bootstrapped samples with only one label. For scores like AUC we
    need at least one positive and one negative sample. (default: True)
    :return: Computed p-value, array of bootstrapped differences of scores.
    """

    y_true = np.array(y_true)
    y_preds1 = np.atleast_2d(y_preds1)
    y_preds2 = np.atleast_2d(y_preds2)
    assert all(len(y_true) == len(y) for y in y_preds1)
    assert all(len(y_true) == len(y) for y in y_preds2)

    np.random.seed(seed)
    z = []
    for i in range(n_bootstraps):
        readers1 = np.random.randint(0, len(y_preds1), len(y_preds1))
        readers2 = np.random.randint(0, len(y_preds2), len(y_preds2))
        indices = np.random.randint(0, len(y_true), len(y_true))
        if reject_one_class_samples and len(np.unique(y_true[indices])) < 2:
            continue
        reader1_scores = []
        for r in readers1:
            if sample_weight is not None:
                reader1_scores.append(score_fun(y_true[indices], y_preds1[r][indices], sample_weight=sample_weight[indices]))
            else:
                reader1_scores.append(score_fun(y_true[indices], y_preds1[r][indices]))
        score1 = stat_fun(reader1_scores)
        reader2_scores = []
        for r in readers2:
            if sample_weight is not None:
                reader2_scores.append(score_fun(y_true[indices], y_preds2[r][indices], sample_weight=sample_weight[indices]))
            else:
                reader2_scores.append(score_fun(y_true[indices], y_preds2[r][indices]))
        score2 = stat_fun(reader2_scores)
        z.append(compare_fun(score1, score2))

    p = percentileofscore(z, 0.0, kind="weak") / 100.0
    if two_tailed:
        p *= 2.0
    return p, z



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict, Tuple

import numpy as np
from scipy import ndimage
from sklearn.metrics import auc

try:
    import numpy.typing as npt
except ImportError:  # pragma: no cover
    pass


label_structure = np.ones((3, 3, 3))


# parse detection maps to individual lesions + confidences
def parse_detection_map(
    y_det: "npt.NDArray[np.float32]"
) -> "Tuple[Dict[int, float], npt.NDArray[np.int32]]":
    """Extract confidence scores per lesion candidate"""
    # label all non-connected components in the detection map
    blobs_index, num_blobs = ndimage.label(y_det, structure=label_structure)

    # input verification
    if num_blobs < len(set(np.unique(y_det))-{0}):
        raise ValueError(
            "It looks like you provided your predictions as softmax volumes instead of detection maps. "
            "If this is the case, please convert your softmax volumes to detection maps (e.g. via postprocessing). "
            "Visit https://github.com/DIAGNijmegen/picai_eval/ for documentation on what we expect of `detection maps` "
            "and how they can be generated from softmax predictions."
        )

    # extract confidence per lesion candidate
    confidences = {}
    for lesion_candidate_id in range(num_blobs):
        max_prob = y_det[blobs_index == (1+lesion_candidate_id)].max()
        confidences[lesion_candidate_id] = float(max_prob)

    return confidences, blobs_index


def calculate_dsc(y_det: "npt.NDArray[np.float32]", y_true: "npt.NDArray[np.int32]") -> float:
    """Calculate Dice similarity coefficient (DSC) for N-D Arrays"""
    epsilon = 1e-8
    dsc_num = np.sum(y_det[y_true == 1]) * 2.0
    dsc_denom = np.sum(y_det) + np.sum(y_true)
    return float((dsc_num + epsilon) / (dsc_denom + epsilon))


def calculate_iou(y_det: "npt.NDArray[np.float32]", y_true: "npt.NDArray[np.int32]") -> float:
    """Calculate Intersection over Union (IoU) for N-D Arrays"""
    epsilon = 1e-8
    iou_num = np.sum(y_det[y_true == 1])
    iou_denom = np.sum(y_det) + np.sum(y_true) - iou_num
    return float((iou_num + epsilon) / (iou_denom + epsilon))


def calculate_operating_points(y, x, op_match=None, verbose=1):
    """
    Calculate operating points for a curve.

    Input:
    - y: (monotonically increasing) performance metric, such as the True Positive Rate
    - x: (monotonically increasing) performance metric, such as the False Positive Rate
    - op_match: dictionary that specifies the target operating point: {
        'x': target x value, 'y': target y value
    }

    Returns:
    - dictionary with operating point(s): {
        'op_closest_xy_y': y_op, # y value at operating point that matches both x and y of target operating point
        'op_closest_xy_x': x_op, # x value at operating point that matches both x and y of target operating point
        ...
    }
    """
    # TODO: currently, a lower sensitivity is preferrred over a higher sensitivity if that means the operating point is matched better.
    # Would be better to go for the best sensitivity/specificity, if that can be done without hurting the other performance metric.
    # In practice, this should not be an issue, as we have many points then.
    y = np.array(y)
    x = np.array(x)
    operating_points = {}

    if not np.all(np.diff(y) >= 0) and verbose:
        print("Warning: y performance metric is not monotonically increasing, this could lead to unexpected behaviour!")
    if not np.all(np.diff(x) >= 0) and verbose:
        print("Warning: x performance metric is not monotonically increasing, this could lead to unexpected behaviour!")

    # grab index of intersection -> compute y/TPR and x/FPR @ index -> store
    op_best_roc_idx = np.argmin(np.abs(y - (1 - x)))
    op_best_roc_y = y[op_best_roc_idx]
    op_best_roc_x = x[op_best_roc_idx]
    operating_points.update(dict(
        op_best_roc_idx=op_best_roc_idx,
        op_best_roc_y=op_best_roc_y,
        op_best_roc_x=op_best_roc_x
    ))

    if op_match is not None:
        # calculate operating point closest to target operating point
        abs_deficit_x, abs_deficit_y = None, None
        optional_x_keys = ['x', 'fpr', 'FPR']
        optional_y_keys = ['y', 'tpr', 'TPR', 'sensitivity', 'sens']

        # if the target x value is specified, calculate the difference between target and oberved value
        for key in optional_x_keys:
            if key in op_match:
                op_match_x = op_match[key]
                abs_deficit_x = np.abs(x - op_match_x)
                break

        # if the target y value is specified, calculate the difference between target and oberved value
        for key in optional_y_keys:
            if key in op_match:
                op_match_y = op_match[key]
                abs_deficit_y = np.abs(y - op_match_y)
                break

        # if both target x and y values are specified, calculate the difference between the target pair and observed pair
        # at the best match, store the observed x and y values
        if abs_deficit_x is not None and abs_deficit_y is not None:
            # determine the index of the the closest point to the target pair
            abs_deficit = abs_deficit_x + abs_deficit_y
            op_closest_xy_idx = np.argmin(abs_deficit)
            op_closest_xy_y = y[op_closest_xy_idx]
            op_closest_xy_x = x[op_closest_xy_idx]

            # store
            operating_points.update(dict(
                op_closest_xy_idx=op_closest_xy_idx,
                op_closest_xy_y=op_closest_xy_y,
                op_closest_xy_x=op_closest_xy_x
            ))

        # same for matching x only
        if abs_deficit_x is not None:
            # determine the index of the the closest point to the target value
            op_closest_x_idx = np.argmin(abs_deficit_x)
            op_closest_x_y = y[op_closest_x_idx]
            op_closest_x_x = x[op_closest_x_idx]

            # store
            operating_points.update(dict(
                op_closest_x_idx=op_closest_x_idx,
                op_closest_x_y=op_closest_x_y,
                op_closest_x_x=op_closest_x_x
            ))

        # same for matching y only
        if abs_deficit_y is not None:
            # determine the index of the the closest point to the target value
            op_closest_y_idx = np.argmin(abs_deficit_y)
            op_closest_y_y = y[op_closest_y_idx]
            op_closest_y_x = x[op_closest_y_idx]

            # store
            operating_points.update(dict(
                op_closest_y_idx=op_closest_y_idx,
                op_closest_y_y=op_closest_y_y,
                op_closest_y_x=op_closest_y_x
            ))

    return operating_points


# calculate statistics for multiple curves
def calculate_statistics(metrics, op_match=None, x_start=0., x_end=1., verbose=1):
    """
    Calculate statistics, such as the area under the curve, for multiple (independent) curves.
    To calculate shared statistics, the curves must be translated to a shared x domain. To
    achieve this with virtually no loss of the step-like nature of curves like ROC and FROC,
    the shared x values are derived from the input, and offset with ± 1e-7.

    Input:
    - metrics should be a list of tuples with the y & x coordinates for each run:
    [([y1, y2, y3, ...], [x1, x2, x3]), # run 1
     ([y1, y2, y3, ...], [x1, x2, x3]), # run 2
     ]
    - op_match: {
        'y': value of y metric (e.g., TPR/sensitivity) to match,
        'x': value of x metric (e.g., FPR/false positive rate) to match,
    }

    Note: mean and 95% CI are calculated as function of the shared x.
    """
    # construct the array of shared x values
    eps = 1e-10
    x_shared = np.array([xi for _, x in metrics for xi in x], dtype=np.float64)  # collect list of all possible x-values
    x_shared = np.ravel(x_shared)  # flatten list, if necessary
    x_shared = np.append(x_shared, [x_start, x_end])  # add x_start and x_end to ensure correct pAUC calculation
    x_shared = np.concatenate((x_shared+eps, x_shared-eps))
    x_shared = np.unique(x_shared)  # only keep unique x values
    x_shared.sort()  # sort in ascending order (inplace)

    # validate x_start and x_end
    assert x_start < x_end, f"x_start must be smaller than x_end! Got x_start={x_start} and x_end={x_end}."

    # convert the per-model y (e.g., TPR) vs x (e.g., FPR) to a shared domain
    y_shared_all = np.zeros(shape=(len(metrics), len(x_shared)), dtype=np.float32)
    auroc_all = []
    individually_matched_operating_points = []
    for i, (y, x) in enumerate(metrics):
        # if necessary, unpack x and y
        if len(y) == 1:
            y = y[0]
        if len(x) == 1:
            x = x[0]

        # interpolate the y values to the shared x values
        y_shared_domain = np.interp(x_shared, x, y)
        y_shared_all[i] = y_shared_domain

        # calculate AUROC for macro stats
        mask = (x_shared >= x_start) & (x_shared <= x_end)
        auc_score = auc(x_shared[mask], y_shared_domain[mask])
        auroc_all += [auc_score]

        # match operating point for each run individually
        operating_points = calculate_operating_points(y=y, x=x, op_match=op_match)
        individually_matched_operating_points += [operating_points]

    # calculate statistics in shared domain
    y_shared_mean = np.mean(y_shared_all, axis=0)
    y_shared_std = np.std(y_shared_all, axis=0)
    y_shared_CI_lower = np.percentile(y_shared_all, 2.5, axis=0)
    y_shared_CI_higher = np.percentile(y_shared_all, 97.5, axis=0)
    auroc_mean = np.mean(auroc_all)
    auroc_std = np.std(auroc_all)

    # calculate operating points in shared domain
    operating_points = calculate_operating_points(y=y_shared_mean, x=x_shared,
                                                  op_match=op_match, verbose=verbose)

    # collect results
    results = {
        # overview statistics
        'auroc_mean': auroc_mean,
        'auroc_std': auroc_std,
        'auroc_all': auroc_all,

        # for plotting
        'x_shared': x_shared,
        'y_shared_all': y_shared_all,
        'y_shared_mean': y_shared_mean,
        'y_shared_std': y_shared_std,
        'y_shared_CI_lower': y_shared_CI_lower,
        'y_shared_CI_higher': y_shared_CI_higher,

        # individually matched operating point
        'individually_matched_operating_points': individually_matched_operating_points,
    }
    results.update(operating_points)

    # calculate standard deviation of each metric (op_closest_xy_y, etc.) between individual runs
    individually_matched_operating_points_std = {
        f"{key}_std": np.std([
            operating_point_info[key]
            for operating_point_info in individually_matched_operating_points
        ])
        for key in individually_matched_operating_points[0].keys()
    }
    results.update(individually_matched_operating_points_std)

    return results


def calculate_pAUC_from_graph(x, y, pAUC_start: float = 0.0, pAUC_end: float = 1.0, full: bool = False):
    """
    Calculate (partial) Area Under Curve (pAUC) using (x,y) coordinates from the given curve.

    Input:
    For a single curve:
    - x: x values of a curve (e.g., the False Positive Rate points). [x1, x2, .., xn]
    - y: y values of a curve (e.g., the True Positive Rate points). [y1, y2, .., yn]

    For multiple curves:
    - list of x curves, for example the x values observed across multiple runs. [[x1m1, x2m1, .., xnm1], [x1m2, x2m2, ...., xnm2], ..]
    - list of y curves, for example the y values observed across multiple runs. [[y1m1, y2m1, .., ynm1], [y1m2, y2m2, ...., ynm2], ..]

    - pAUC_start: lower bound of x (e.g., FPR) to compute pAUC
    - pAUC_end: higher bound of x (e.g., FPR) to compute pAUC

    Returns:
    - if (full==False): List of pAUC values for each set of ([x1, ..], [y1, ..]) coordinates
    - if (full==True): Metrics as returned by `calculate_statistics` [see there]

    Note: function is not specific to the FROC curve
    """

    if not isinstance(x[0], (list, np.ndarray)) or not isinstance(y[0], (list, np.ndarray)):
        # have a single set of (x,y) coordinates

        assert not isinstance(x[0], (list, np.ndarray)) and not isinstance(y[0], (list, np.ndarray)), \
            "Either provide multiple sequences of (x,y) coordinates, or a single sequence. Obtained a mix of both now. "

        # pack coordinates in format expected by `calculate_statistics`
        coordinates_joined = [(y, x)]
    else:
        # have multiple sets of (x,y) coordinates
        # pack coordinates in format expected by `calculate_statistics`
        coordinates_joined = list(zip(y, x))

    # calculate AUC in the given range
    results = calculate_statistics(metrics=coordinates_joined, x_start=pAUC_start, x_end=pAUC_end)
    if full:
        return results
    return results['auroc_all']



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division, print_function

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Hashable, List, Optional, Tuple, Union

import numpy as np
from sklearn.metrics import auc, precision_recall_curve, roc_curve

try:
    import numpy.typing as npt
except ImportError:  # pragma: no cover
    pass

from picai_eval.data_utils import PathLike, load_metrics, save_metrics


@dataclass
class Metrics:
    lesion_results: Union[Dict[Hashable, List[Tuple[int, float, float]]], PathLike]
    case_target: Optional[Dict[Hashable, int]] = None
    case_pred: Optional[Dict[Hashable, float]] = None
    case_weight: Optional[Union[Dict[Hashable, float], List[float]]] = None
    lesion_weight: Optional[Dict[Hashable, List[float]]] = None
    thresholds: "Optional[npt.NDArray[np.float64]]" = None
    subject_list: Optional[List[str]] = None
    sort: bool = True

    def __post_init__(self):
        if isinstance(self.lesion_results, (str, Path)):
            # load metrics from file
            self.load(self.lesion_results)

        if self.subject_list is None:
            self.subject_list = sorted(list(self.lesion_results))

        if self.case_target is None:
            # derive case-level targets as the maximum lesion-level target
            self.case_target = {
                idx: max([is_lesion for is_lesion, _, _ in case_y_list]) if len(case_y_list) else 0
                for idx, case_y_list in self.lesion_results.items()
            }

        if self.case_pred is None:
            # derive case-level predictions as the maximum lesion-level prediction
            self.case_pred = {
                idx: max([confidence for _, confidence, _ in case_y_list]) if len(case_y_list) else 0
                for idx, case_y_list in self.lesion_results.items()
            }

        if not isinstance(self.case_weight, dict):
            subject_list = list(self.case_target)
            if self.case_weight is None:
                self.case_weight = {idx: 1 for idx in subject_list}
            else:
                self.case_weight = {idx: weight for idx, weight in zip(subject_list, self.case_weight)}

        if self.lesion_weight is None:
            subject_list = sorted(list(self.lesion_results))
            self.lesion_weight = {idx: [1]*len(case_y_list) for idx, case_y_list in self.lesion_results.items()}

        if self.sort:
            # sort dictionaries
            subject_list = sorted(list(self.lesion_results))
            self.lesion_results = {idx: self.lesion_results[idx] for idx in subject_list}
            self.lesion_weight = {idx: self.lesion_weight[idx] for idx in subject_list}
            self.case_target = {idx: self.case_target[idx] for idx in subject_list}
            self.case_pred = {idx: self.case_pred[idx] for idx in subject_list}
            self.case_weight = {idx: self.case_weight[idx] for idx in subject_list}

    # aggregates
    def calc_auroc(self, subject_list: Optional[List[str]] = None) -> float:
        """Calculate case-level Area Under the Receiver Operating Characteristic curve (AUROC)"""
        return self.calculate_ROC(subject_list=subject_list)['AUROC']

    @property
    def auroc(self) -> float:
        """Calculate case-level Area Under the Receiver Operating Characteristic curve (AUROC)"""
        return self.calc_auroc()

    def calc_AP(self, subject_list: Optional[List[str]] = None) -> float:
        """Calculate Average Precision"""
        return self.calculate_precision_recall(subject_list=subject_list)['AP']

    @property
    def AP(self) -> float:
        """Calculate Average Precision"""
        return self.calc_AP()

    @property
    def num_cases(self) -> int:
        """Calculate the number of cases"""
        return len(self.subject_list)

    @property
    def num_lesions(self) -> int:
        """Calculate the number of ground truth lesions"""
        return sum([is_lesion for is_lesion, *_ in self.lesion_results_flat])

    @property
    def score(self):
        """Calculate the ranking score, as used in the PI-CAI 22 Grand Challenge"""
        return (self.auroc + self.AP) / 2

    # lesion-level results
    def get_lesion_results_flat(self, subject_list: Optional[List[str]] = None):
        """Flatten the per-case lesion evaluation results into a single list"""
        if subject_list is None:
            subject_list = self.subject_list

        return [
            (is_lesion, confidence, overlap)
            for subject_id in subject_list
            for is_lesion, confidence, overlap in self.lesion_results[subject_id]
        ]

    @property
    def lesion_results_flat(self) -> List[Tuple[int, float, float]]:
        """Flatten the per-case y_list"""
        return self.get_lesion_results_flat()

    def get_lesion_weight_flat(self, subject_list: Optional[List[str]] = None) -> List[float]:
        """Retrieve lesion-wise sample weights (for a given subset of cases)"""
        if subject_list is None:
            subject_list = self.subject_list

        # collect lesion weights (and flatten)
        return [weight for subject_id in subject_list for weight in self.lesion_weight[subject_id]]

    @property
    def lesion_weight_flat(self) -> List[float]:
        """Retrieve lesion-wise sample weights (for a given subset of cases)"""
        return self.get_lesion_weight_flat()

    @property
    def precision(self) -> "npt.NDArray[np.float64]":
        """Calculate lesion-level precision at each threshold"""
        return self.calculate_precision_recall()['precision']

    @property
    def recall(self) -> "npt.NDArray[np.float64]":
        """Calculate lesion-level recall at each threshold"""
        return self.calculate_precision_recall()['recall']

    @property
    def lesion_TP(self) -> "npt.NDArray[np.float64]":
        """Calculate number of true positive lesion detections at each threshold"""
        return self.calculate_counts()['TP']

    @property
    def lesion_FP(self) -> "npt.NDArray[np.float64]":
        """Calculate number of false positive lesion detections at each threshold"""
        return self.calculate_counts()['FP']

    @property
    def lesion_TPR(self) -> "npt.NDArray[np.float64]":
        """Calculate lesion-level true positive rate (sensitivity) at each threshold"""
        if self.num_lesions > 0:
            return self.lesion_TP / self.num_lesions
        else:
            return np.array([np.nan] * len(self.lesion_TP))

    @property
    def lesion_FPR(self) -> "npt.NDArray[np.float64]":
        """Calculate lesion-level false positive rate (number of false positives per case) at each threshold"""
        return self.lesion_FP / self.num_cases

    # case-level results
    def calc_case_TPR(self, subject_list: Optional[List[str]] = None) -> "npt.NDArray[np.float64]":
        """Calculate case-level true positive rate (sensitivity) at each threshold"""
        return self.calculate_ROC(subject_list=subject_list)['TPR']

    @property
    def case_TPR(self) -> "npt.NDArray[np.float64]":
        """Calculate case-level true positive rate (sensitivity) at each threshold"""
        return self.calc_case_TPR()

    def calc_case_FPR(self, subject_list: Optional[List[str]] = None) -> "npt.NDArray[np.float64]":
        """Calculate case-level false positive rate (1 - specificity) at each threshold"""
        return self.calculate_ROC(subject_list=subject_list)['FPR']

    @property
    def case_FPR(self) -> "npt.NDArray[np.float64]":
        """Calculate case-level false positive rate (1 - specificity) at each threshold"""
        return self.calc_case_FPR()

    # supporting functions
    def calculate_counts(self, subject_list: Optional[List[str]] = None) -> "Dict[str, npt.NDArray[np.float32]]":
        """
        Calculate lesion-level true positive (TP) detections and false positive (FP) detections as each threshold.
        """
        # flatten y_list (and select cases in subject_list)
        lesion_y_list = self.get_lesion_results_flat(subject_list=subject_list)

        # collect targets and predictions
        y_true: "npt.NDArray[np.float64]" = np.array([target for target, *_ in lesion_y_list])
        y_pred: "npt.NDArray[np.float64]" = np.array([pred for _, pred, *_ in lesion_y_list])

        if self.thresholds is None:
            # collect thresholds for lesion-based analysis
            self.thresholds = np.unique(y_pred)
            self.thresholds[::-1].sort()  # sort thresholds in descending order (inplace)

            # for >10,000 thresholds: resample to 10,000 unique thresholds, while also
            # keeping all thresholds higher than 0.8 and the first 20 thresholds
            if len(self.thresholds) > 10_000:
                rng = np.arange(1, len(self.thresholds), len(self.thresholds)/10_000, dtype=np.int32)
                st = [self.thresholds[i] for i in rng]
                low_thresholds = self.thresholds[-20:]
                self.thresholds = np.array([t for t in self.thresholds if t > 0.8 or t in st or t in low_thresholds])

        # define placeholders
        FP: "npt.NDArray[np.float32]" = np.zeros_like(self.thresholds, dtype=np.float32)
        TP: "npt.NDArray[np.float32]" = np.zeros_like(self.thresholds, dtype=np.float32)

        # for each threshold: count FPs and TPs
        for i, th in enumerate(self.thresholds):
            y_pred_thresholded = (y_pred >= th).astype(int)
            tp = np.sum(y_true*y_pred_thresholded)
            fp = np.sum(y_pred_thresholded - y_true*y_pred_thresholded)

            # update with new point
            FP[i] = fp
            TP[i] = tp

        # extend curve to infinity
        TP[-1] = TP[-2]
        FP[-1] = np.inf

        return {
            'TP': TP,
            'FP': FP,
        }

    def calculate_precision_recall(self, subject_list: Optional[List[str]] = None) -> Dict[str, Any]:
        """
        Generate Precision-Recall curve and calculate average precision (AP).
        """
        # flatten y_list (and select cases in subject_list)
        lesion_y_list = self.get_lesion_results_flat(subject_list=subject_list)

        # collect targets and predictions
        y_true: "npt.NDArray[np.float64]" = np.array([target for target, *_ in lesion_y_list])
        y_pred: "npt.NDArray[np.float64]" = np.array([pred for _, pred, *_ in lesion_y_list])

        # calculate precision-recall curve
        precision, recall, thresholds = precision_recall_curve(
            y_true=y_true,
            probas_pred=y_pred,
            sample_weight=self.get_lesion_weight_flat(subject_list=subject_list)
        )

        # set precision to zero at a threshold of "zero", as those lesion
        # candidates are included just to convey the number of lesions to
        # the precision_recall_curve function, and not actual candidates
        precision[:-1][thresholds == 0] = 0

        # calculate average precision using the step function integral
        # The following works because the last entry of precision is
        # guaranteed to be 1, as returned by precision_recall_curve
        # Taken from https://github.com/scikit-learn/scikit-learn/blob/
        # 32f9deaaf27c7ae56898222be9d820ba0fd1054f/sklearn/metrics/_ranking.py#L212
        AP = -np.sum(np.diff(recall) * np.array(precision)[:-1])

        return {
            'AP': AP,
            'precision': precision,
            'recall': recall,
        }

    def calculate_ROC(self, subject_list: Optional[List[str]] = None) -> Dict[str, Any]:
        """
        Generate Receiver Operating Characteristic curve for case-level risk stratification.
        """
        if subject_list is None:
            subject_list = self.subject_list

        fpr, tpr, _ = roc_curve(
            y_true=[self.case_target[s] for s in subject_list],
            y_score=[self.case_pred[s] for s in subject_list],
            sample_weight=[self.case_weight[s] for s in subject_list],
        )

        auroc = auc(fpr, tpr)

        return {
            'FPR': fpr,
            'TPR': tpr,
            'AUROC': auroc,
        }

    @property
    def version(self):
        return "1.4.x"

    def as_dict(self):
        return {
            # aggregates
            "auroc": self.auroc,
            "AP": self.AP,
            "num_cases": self.num_cases,
            "num_lesions": self.num_lesions,
            "picai_eval_version": self.version,

            # lesion-level results
            "lesion_results": self.lesion_results,
            "lesion_weight": self.lesion_weight,

            # case-level results
            "case_pred": self.case_pred,
            "case_target": self.case_target,
            "case_weight": self.case_weight,
        }

    def full_dict(self):
        return {
            # aggregates
            "auroc": self.auroc,
            "AP": self.AP,
            "num_cases": self.num_cases,
            "num_lesions": self.num_lesions,
            "picai_eval_version": self.version,

            # lesion-level results
            "lesion_results": self.lesion_results,
            "lesion_weight": self.lesion_weight,
            "precision": self.precision,
            "recall": self.recall,
            'lesion_TPR': self.lesion_TPR,
            'lesion_FPR': self.lesion_FPR,
            "thresholds": self.thresholds,

            # case-level results
            "case_pred": self.case_pred,
            "case_target": self.case_target,
            "case_weight": self.case_weight,
        }

    def minimal_dict(self):
        return {
            # lesion-level results
            "lesion_results": self.lesion_results,
            "lesion_weight": self.lesion_weight,

            # case-level results
            "case_pred": self.case_pred,
            "case_target": self.case_target,
            "case_weight": self.case_weight,
        }

    def save(self, path: PathLike):
        """Save metrics to file (including aggregates)"""
        save_metrics(metrics=self.as_dict(), file_path=path)

    def save_full(self, path: PathLike):
        """Save metrics to file (including derived metrics)"""
        save_metrics(metrics=self.full_dict(), file_path=path)

    def save_minimal(self, path: PathLike):
        """Save metrics to file (minimal required metrics)"""
        save_metrics(metrics=self.minimal_dict(), file_path=path)

    def load(self, path: PathLike):
        """Load metrics from file"""
        metrics = load_metrics(path)

        # parse metrics
        self.case_target = {idx: int(float(val)) for idx, val in metrics['case_target'].items()}
        self.case_pred = {idx: float(val) for idx, val in metrics['case_pred'].items()}
        self.case_weight = {idx: float(val) for idx, val in metrics['case_weight'].items()}
        self.lesion_weight = {idx: [float(val) for val in weights] for idx, weights in metrics['lesion_weight'].items()}
        self.lesion_results = {
            idx: [
                (int(float(is_lesion)), float(confidence), float(overlap))
                for (is_lesion, confidence, overlap) in lesion_results_case
            ]
            for idx, lesion_results_case in metrics['lesion_results'].items()
        }

    def __str__(self) -> str:
        return f"Metrics(auroc={self.auroc:.2%}, AP={self.AP:.2%}, {self.num_cases} cases, {self.num_lesions} lesions)"

    def __repr__(self) -> str:
        return self.__str__()



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import concurrent.futures
from collections.abc import Sequence
from itertools import repeat
from multiprocessing import Pool
from typing import Any, Callable, Dict, Hashable, List, Optional, Union

import numpy as np
from mlxtend.evaluate import permutation_test
from sklearn.metrics import (cohen_kappa_score, confusion_matrix,
                             roc_auc_score, roc_curve)
from tqdm import tqdm

try:
    from picai_eval.stat_util.stat_util import pvalue_stat
except ImportError:  # pragma: no cover
    print("Could not find picai_eval.stat_util. Please install picai_eval including submodules (use pip >= 21.0)")

try:
    import numpy.typing as npt
except ImportError:  # pragma: no cover
    pass

"""
Binary PCa Detection in mpMRI
Script:         Statistical helper - (non-parametric) statistical test
Contributor:    joeranbosma
Target Organ:   Prostate
Target Classes: Benign(0), Malignant(1)
"""

IntDictOrArraylike = "Union[Dict[Hashable, int], Sequence[int], npt.NDArray[np.int_]]"
FloatDictOrArraylike = "Union[Dict[Hashable, float], Sequence[float], npt.NDArray[np.float_]]"


def calc_sensitivity(y_true, y_pred, sample_weight=None):
    """Calculate sensitivity"""
    _, _, fn, tp = confusion_matrix(y_true, y_pred, sample_weight=sample_weight).ravel()
    return tp / (tp+fn)


def calc_specificity(y_true, y_pred, sample_weight=None):
    """Calculate specificity"""
    tn, fp, _, _ = confusion_matrix(y_true, y_pred, sample_weight=sample_weight).ravel()
    return tn / (tn+fp)


def calc_order_auc_score(
    scores_baseline: "Union[Sequence[float], npt.NDArray[np.float_]]",
    scores_alternative: "Union[Sequence[float], npt.NDArray[np.float_]]"
) -> float:
    """Calculate the order AUC for performance(alternative) > performance(baseline)"""
    # pose order AUC as classification AUC to use roc_auc_score
    scores_all = list(scores_baseline) + list(scores_alternative)
    labels_all = [0]*len(list(scores_baseline)) + [1]*len(list(scores_alternative))
    auc_score = roc_auc_score(y_true=labels_all, y_score=scores_all)

    return float(auc_score)


def perform_permutation_test(
    scores_baseline: "Union[Sequence[float], npt.NDArray[np.float_]]",
    scores_alternative: "Union[Sequence[float], npt.NDArray[np.float_]]",
    method: Optional[str] = None,
    iterations: int = 1_000_000
) -> float:
    """
    Calculate p-value for comparison of baseline with alternative configuration.
    H0 (null hypothesis): difference in performance comes from random fluctuation
    H1 (alternative hypothesis): alternative configuration is better

    Definitions:
    - configuration: either baseline or alternative
    - performance: metric for performance (e.g., AUC, pAUC between 0.01 and 2.5, accuracy, ...), where higher is better.
    - run: a single fully trained model. Having 3 restarts and 5-fold cross-validation will give 15 runs.

    Usage:
    Collect the performances observed for each run in lists:
    - scores_baseline: [performance baseline run 1, performance baseline run 2, ..., performance baseline run N]
    - scores_alternative: [performance alternative run 1, performance alternative run 2, ..., performance alternative run M]

    - iterations: number of permuations to perform to estimate the distribution of the null hypothesis (H0)

    Returns:
    - p: p-value for null hypothesis (so alternative configuration is better than the baseline if p is sufficiently small)
    """
    if method is None:
        # use exact permutation test if feasible, or approximate permutation test otherwise
        method = 'exact' if len(scores_baseline)+len(scores_alternative) <= 20 else 'approximate'

    # perform permutation test
    p = permutation_test(
        x=scores_baseline, y=scores_alternative, func=calc_order_auc_score,
        method=method, num_rounds=iterations
    )

    return float(p)


def permutation_test_wrapper(args) -> float:
    return perform_permutation_test(**args)


def multiple_permutation_tests(
    all_scores: "Dict[str, Dict[str, Union[Sequence[float], npt.NDArray[np.float_]]]]",
    iterations: int = 1_000_000
) -> Dict[str, float]:
    """
    Perform multiple permutation tests using multiprocessing.

    Input:
    - scores: {
        'name of comparison 1': {
            'scores_baseline': [baseline scores comparison 1 (see `perform_permutation_test`)],
            'scores_alternative': [alternative scores comparison 1 (see `perform_permutation_test`)],
        },
        'name of comparison 2': {
            'scores_baseline': [baseline scores comparison 2 (see `perform_permutation_test`)],
            'scores_alternative': [alternative scores comparison 2 (see `perform_permutation_test`)],
        },
        ...,
        'name of comparison n': {
            'scores_baseline': [baseline scores comparison n (see `perform_permutation_test`)],
            'scores_alternative': [alternative scores comparison n (see `perform_permutation_test`)],
        },
    }

    Example scores:

    Returns:
    - p_values: {
        'name of comparison 1': p-value for H0 ≥ H1 of comparison 1 (see `perform_permutation_test`),
        'name of comparison 2': p-value for H0 ≥ H1 of comparison 2 (see `perform_permutation_test`),
        ...,
        'name of comparison n': p-value for H0 ≥ H1 of comparison n (see `perform_permutation_test`),
    }
    """
    # define placeholder
    p_values: Dict[str, float] = {}

    with concurrent.futures.ProcessPoolExecutor() as pool:
        jobs = {}
        for name, scores in all_scores.items():
            # prepare input arguments of permutation test
            args: Dict[str, Any] = dict(scores)
            args['iterations'] = iterations

            # add permutation test to the queue
            future = pool.submit(permutation_test_wrapper, args)
            jobs[future] = name

        for future in tqdm(concurrent.futures.as_completed(jobs), total=len(scores), desc="Performing permutation tests"):
            # store result of permutation test
            name = jobs[future]
            p_values[name] = future.result()

    return p_values


def match_reader(y_true, y_pred_ai, y_pred_reader, sample_weight=None, match='sensitivity'):
    # input conversion and validation
    y_true = np.array(y_true)
    y_pred_ai = np.array(y_pred_ai)
    y_pred_reader = np.array(y_pred_reader)
    assert len(y_true) == len(y_pred_reader)

    if y_true.shape != y_pred_ai.shape:
        return np.array([
            match_reader(
                y_true=y_true,
                y_pred_ai=y_pred_ai_restart,
                y_pred_reader=y_pred_reader,
                sample_weight=sample_weight,
                match=match
            )
            for y_pred_ai_restart in y_pred_ai
        ])

    assert len(y_true) == len(y_pred_ai)

    # calculate metric (sensitivity/specificity/recall/precision) for reader
    if match == 'sensitivity':
        performance_reader = calc_sensitivity(y_true=y_true, y_pred=y_pred_reader, sample_weight=sample_weight)
    elif match == 'specificity':
        performance_reader = calc_specificity(y_true=y_true, y_pred=y_pred_reader, sample_weight=sample_weight)
    else:
        raise ValueError

    # match AI operating point to reader
    fpr, tpr, thresholds = roc_curve(y_true=y_true, y_score=y_pred_ai, sample_weight=sample_weight)
    if match == 'sensitivity':
        performance_ai = tpr
    elif match == 'specificity':
        performance_ai = 1 - fpr

    diff = np.abs(performance_ai - performance_reader)

    # grab index of closest point
    viable_thresholds = thresholds[diff == np.min(diff)]
    threshold = np.random.choice(viable_thresholds)
    # closest_idx = np.argmin(diff)
    # threshold = thresholds[closest_idx]

    return (y_pred_ai >= threshold).astype(int)


def match_then_compare(y_true, y_pred_ai, y_pred_reader, sample_weight=None, match='sensitivity', verbose=False):
    # match AI operating point to the reader
    y_pred_ai_matched = match_reader(
        y_true=y_true,
        y_pred_ai=y_pred_ai,
        y_pred_reader=y_pred_reader,
        sample_weight=sample_weight,
        match=match
    )

    # evaluate AI and reader at matched operating point
    kappa_ai = cohen_kappa_score(y1=y_true, y2=y_pred_ai_matched, sample_weight=sample_weight)
    kappa_reader = cohen_kappa_score(y1=y_true, y2=y_pred_reader, sample_weight=sample_weight)

    print(f"κR = {kappa_reader:.4f}, κAI = {kappa_ai:.4f}") if verbose else None

    # return test statistic
    return kappa_reader - kappa_ai


def sample_then_match_then_compare(y_true, y_pred_ai, y_pred_reader, sample_weight, match='sensitivity'):
    indices = np.random.randint(0, len(y_true), len(y_true))
    while len(np.unique(y_true[indices])) < 2:
        # Reject one class samples
        indices = np.random.randint(0, len(y_true), len(y_true))

    z = match_then_compare(
        y_true=y_true[indices],
        y_pred_ai=y_pred_ai[indices],
        y_pred_reader=y_pred_reader[indices],
        sample_weight=sample_weight[indices],
        match=match
    )

    return z


def sample_then_match_then_compare_wrapper(args):  # pragma: no cover
    return sample_then_match_then_compare(**args)


def validate_and_convert(*inputs) -> "List[npt.NDArray[Any]]":
    """
    Validate inputs:
    - If the inputs consists of at least one dictionary, all inputs should be a dictionary
    - If the inputs are dictionaries, check if the keys are the same

    Convert inputs:
    - If the inputs are dictionaries, reduce to flat lists, with the same order for each input
    - Convert to numpy arrays
    """
    # check if any of the inputs is a dictionary
    any_dict = False
    for inp in inputs:
        if isinstance(inp, dict):
            any_dict = True

    if any_dict:
        # check if all inputs are dictionaries
        assert all(isinstance(inp, dict) for inp in inputs), (
            "Inputs must either all be dictionaries, or all "
            "iterables (with cases in the same order)!"
        )

        # check if all cases are present in each dictionary
        cases = set(list(inputs[0]))
        assert all(cases == set(list(inp)) for inp in inputs), \
            "Inputs must all contain the same cases!"

        # collect flat lists with cases in the same order
        cases = sorted(list(cases))
        inputs = ([inp[c] for c in cases] for inp in inputs)

    # convert to numpy arrays
    return [np.array(inp) for inp in inputs]


def perform_matched_boostrapping(
    y_true: "Union[Dict[Hashable, int], Sequence[int], npt.NDArray[np.int_]]",
    y_pred_ai: "Union[Dict[Hashable, float], Sequence[float], npt.NDArray[np.float_]]",
    y_pred_reader: "Union[Dict[Hashable, int], Sequence[int], npt.NDArray[np.int_]]",
    match: str = "sensitivity",
    conjugate_performance_metric: Optional[Union[str, Callable[[Any, Any, Optional[Any]], float]]] = None,
    flavour: str = "match>sample>compare",
    sample_weight: "Optional[Union[Dict[Hashable, float], Sequence[float], npt.NDArray[np.float_]]]" = None,
    iterations: int = 1_000_000,
    seed: Optional[int] = None,
    verbose: bool = True
) -> float:
    """
    Perform bootstrapping to estimate the probability that AI outperforms the reader.

    For sample>match>compare, in each iteration of bootstrapping:
    1. select ~U(0, N) samples with replacement
    2. match the AI operating point to the operating point of the reader, by selecting
       the operating point with sensitivity/specificity/recall/precision closest to the
       reader. After matching, the AI predictions are binarized using the selected threshold.
    3. calculate the target metric: reader κ - AI κ
    Iterations that sampled only one class are rejected.

    For match>sample>compare, in each iteration of bootstrapping:
    1. Match the AI operating point to the operating point of the reader, by selecting
       the operating point with sensitivity/specificity/recall/precision closest to the
       reader. After matching, the AI predictions are binarized using the selected threshold.
    2. select ~U(0, N) samples with replacement
    3. calculate the target metric: reader κ - AI κ
    Iterations that sampled only one class are rejected.

    Input:
    - y_true: ground truth containing 0 or 1 for each case, as dictionary (recommended) or list
    - y_pred_ai: continuous predictions for each case, as dictionary (recommended) or list
    - y_pred_reader: binarized predictions for each case, as dictionary (recommended) or list
    - flavour: either "match>sample>compare" or "sample>match>compare"

    Returns:
    - p: estimate of probability for Performance(AI) ≥ Performance(Reader)
    """
    if sample_weight is not None:
        y_true, y_pred_ai, y_pred_reader, sample_weight = validate_and_convert(
            y_true, y_pred_ai, y_pred_reader, sample_weight
        )
    else:
        y_true, y_pred_ai, y_pred_reader = validate_and_convert(
            y_true, y_pred_ai, y_pred_reader
        )

    # convert values
    y_true = y_true.astype(int)
    y_pred_ai = y_pred_ai.astype(float)
    y_pred_reader = y_pred_reader.astype(int)
    assert np.min(y_pred_reader) >= 0 and np.max(y_pred_reader) <= 1, \
        "Reader predictions should be thresholded (e.g., PI-RADS ≥ 4)."
    if sample_weight is not None:
        sample_weight = sample_weight.astype(float)
    else:
        sample_weight = np.ones_like(y_true, dtype=float)

    assert y_true.shape == y_pred_reader.shape and len(y_true.shape) == 1, \
        f"Recieved unexpected shape for y_true (after conversion): {y_true.shape}"

    assert len(y_pred_ai.shape) <= 2, f"Unexpected shape for y_pred_ai (after conversion): {y_pred_ai.shape}"
    if y_pred_ai.shape[-1] != y_true.shape[-1]:
        y_pred_ai = y_pred_ai.transpose()

    if conjugate_performance_metric is None:
        # determine conjugate performance metric
        conjugate_performance_metric = get_conjugate_performance_metric(match)

    if isinstance(conjugate_performance_metric, str):
        # determine conjugate performance metric function
        conjugate_performance_metric = get_performance_metric_func(conjugate_performance_metric)

    if flavour == "match>sample>compare":
        # match AI operating point to the reader
        y_pred_ai_matched = match_reader(
            y_true=y_true,
            y_pred_ai=y_pred_ai,
            y_pred_reader=y_pred_reader,
            sample_weight=sample_weight,
            match=match
        )

        # perform bootstrapping, see https://github.com/mateuszbuda/ml-stat-util
        p, z_subsets = pvalue_stat(
            y_true=y_true,
            y_preds1=y_pred_reader,
            y_preds2=y_pred_ai_matched,
            score_fun=conjugate_performance_metric,
            stat_fun=lambda x: x,
            compare_fun=lambda reader1_scores, reader2_scores: 0.5 - calc_order_auc_score(reader1_scores, reader2_scores),
            sample_weight=sample_weight,
            two_tailed=False,
            n_bootstraps=iterations,
            seed=seed,
        )
    elif flavour == "sample>match>compare":
        with Pool(8) as pool:
            # perform bootstrapping
            # with itertool's repeat the input parameters do not have to be initialized thousands of times
            args = dict(y_true=y_true, y_pred_ai=y_pred_ai, y_pred_reader=y_pred_reader, match=match, sample_weight=sample_weight)
            z_subsets = pool.map(sample_then_match_then_compare_wrapper, repeat(args, iterations))

        # compute probability estimate
        p = np.mean([z < 0 for z in z_subsets])
    else:
        raise ValueError

    if verbose:
        print(f"Probability for Performance(AI) ≥ Performance(Reader): p = {p:.4e}")
        # print(f"z_subsets = {np.mean(z_subsets):.4f} ± {np.std(z_subsets):.4f}")

    return float(p)


def get_conjugate_performance_metric(performance_metric: str):
    # determine conjugate performance metric
    conjugate_performance_metrics = {
        'sensitivity': 'specificity',
        'specificity': 'sensitivity',
        'recall': 'precision',
        'precision': 'recall',
    }
    assert performance_metric in conjugate_performance_metrics, \
        f"Conjugate performance metric for {performance_metric} not found!"

    return conjugate_performance_metrics[performance_metric]


def get_performance_metric_func(performance_metric: str):
    performance_metrics = {
        'sensitivity': calc_sensitivity,
        'specificity': calc_specificity,
        # 'recall': calc_recall,  # TODO: add this
        # 'precision': calc_precision,  # TODO: add this
    }
    assert performance_metric in performance_metrics, \
        f"Implementation of {performance_metric} not found!"

    return performance_metrics[performance_metric]


def perform_matched_permutation_test(
    y_true: IntDictOrArraylike,
    y_pred_ai: FloatDictOrArraylike,
    y_pred_readers: FloatDictOrArraylike,
    match: Union[str, Callable[[Any, Any, Optional[Any]], float]] = "sensitivity",
    conjugate_performance_metric: Optional[Union[str, Callable[[Any, Any, Optional[Any]], float]]] = None,
    sample_weight: Optional[FloatDictOrArraylike] = None,
    iterations: int = 1_000_000,
    verbose: bool = True
) -> float:
    """
    Perform a permutation test to estimate the probability that AI outperforms the panel of readers.

    Input:
    - y_true: ground truth containing 0 or 1 for each case, as dictionary (recommended) or list
    - y_pred_ai: continuous predictions for each case, as dictionary (recommended) or list
    - y_pred_readers: binarized predictions for each case, as dictionary (recommended) or list

    For the readers, the performance metric is obtained in two steps:
        1) for each reader we consider an operating point (PI-RADS 3+ or 4+) and match the trained
           AI models to the reader's sensitivity/specificity/recall/precision,
        2) we calculate the conjugate performance of the reader with respect to the average conjugate
           performance of the AI models. For example, for cased-based analysis, after matching the
           sensitivity, the reader's performance metric is calculated as:
           P_j=spec_R^j-1/N ∑_(i=1)^N〖spec_AI^i 〗, with spec_R^j the reader's specificity and spec_AI^i
           the specificity of trained AI model i at the same sensitivity as the reader.
    This is repeated for each reader, resulting in the performance with respect to the average AI
    performance for each reader.

    To account for variability in AI performance, the same method is applied to each trained AI model:
    we calculate the performance of trained AI models with respect to the average AI performance,
    averaged across the operating points of each reader. For case-based analysis, where sensitivity is
    matched, this evaluates to: P_k=1/M ∑_(j=1)^M(spec_AI^k-1/N ∑_(i=1)^N〖spec_AI^i 〗),  where spec_AI^k
    is the specificity of trained AI model k and spec_AI^i is the specificity of trained AI model i,
    both at the same sensitivity as reader j

    To account for deviation from the population encountered during clinical routine, perform inverse probability
    weighting by providing the weights per sample (see McKinney et al., 2020, Mansournia et al., 2016, Pinsky et al., 2012).

    Returns:
    - p: probability for Performance(AI) ≥ Performance(Panel of readers)
    """
    # validate and convert inputs
    if sample_weight is not None:
        y_true, y_pred_ai, y_pred_readers, sample_weight = validate_and_convert(
            y_true, y_pred_ai, y_pred_readers, sample_weight
        )
    else:
        y_true, y_pred_ai, y_pred_readers = validate_and_convert(
            y_true, y_pred_ai, y_pred_readers
        )

    assert len(y_true.shape) == 1, \
        f"Recieved unexpected shape for y_true (after conversion): {y_true.shape}"

    assert len(y_pred_ai.shape) == 2, f"Unexpected shape for y_pred_ai (after conversion): {y_pred_ai.shape}"
    assert len(y_pred_readers.shape) == 2, f"Unexpected shape for y_pred_readers (after conversion): {y_pred_readers.shape}"
    if y_pred_ai.shape[-1] != y_true.shape[-1]:
        y_pred_ai = y_pred_ai.transpose()
    if y_pred_readers.shape[-1] != y_true.shape[-1]:
        y_pred_readers = y_pred_readers.transpose()

    if conjugate_performance_metric is None:
        # determine conjugate performance metric
        conjugate_performance_metric = get_conjugate_performance_metric(match)

    if isinstance(conjugate_performance_metric, str):
        # determine conjugate performance metric function
        conjugate_performance_metric = get_performance_metric_func(conjugate_performance_metric)

    # keep track of conjugate AI scores, after matching to a reader
    all_scores_ai = []

    # collect performance metrics for readers w.r.t. averge AI
    scores_readers = []
    for y_pred_reader in y_pred_readers:
        # print(f"Sensitivity reader: {calc_sensitivity(y_true, y_pred_reader)}")
        # for each reader, match each AI instance to the reader's performance,
        # and collect the conjugate performance
        matched_conjugate_scores_ai = []
        for y_pred_ai_instance in y_pred_ai:
            y_pred_ai_matched = match_reader(
                y_true=y_true,
                y_pred_ai=y_pred_ai_instance,
                y_pred_reader=y_pred_reader,
                sample_weight=sample_weight,
                match=match
            )

            score_ai = conjugate_performance_metric(y_true, y_pred_ai_matched)
            matched_conjugate_scores_ai.append(score_ai)

        # calculate performance of reader w.r.t. AI and store results
        score_reader = conjugate_performance_metric(y_true, y_pred_reader)
        scores_readers.append(calc_order_auc_score(matched_conjugate_scores_ai, [score_reader]))

        # calculate performance of AI instance w.r.t. AI and store results
        scores_ai = []
        for i, score_ai in enumerate(matched_conjugate_scores_ai):
            other_scores = [score_reader]+matched_conjugate_scores_ai[:i] + matched_conjugate_scores_ai[i+1:]
            scores_ai.append(calc_order_auc_score(other_scores, [score_ai]))
        all_scores_ai.append(scores_ai)

    # collect performance metrics for AI (rank of conjugate performance w.r.t. average AI, averaged across readers)
    scores_ai = np.mean(all_scores_ai, axis=0)

    # perform bootstrapping, see https://github.com/mateuszbuda/ml-stat-util
    print(f"""
    Evaluating:
        {np.mean(scores_readers):.4f} ± {np.std(scores_readers):.4f} for panel of {len(scores_readers)} readers
        {np.mean(scores_ai):.4f} ± {np.std(scores_ai):.4f} for {len(scores_ai)} AI instances
        (with {iterations} iterations)""") if verbose else None

    p = perform_permutation_test(
        scores_baseline=list(scores_readers),
        scores_alternative=list(scores_ai),
        iterations=iterations,
    )

    if verbose:
        print(f"Probability for Performance(Panel of readers) ≥ Performance(AI): p = {p:.4e}")

    return float(p)



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from picai_eval.eval import evaluate, evaluate_folder
from picai_eval.metrics import Metrics

print("If you have questions or suggestions, feel free to open an issue " +
      "at https://github.com/DIAGNijmegen/picai_eval\n")


__all__ = [
    # Explicitly expose these functions for easier imports
    "evaluate",
    "evaluate_folder",
    "Metrics",
]



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import os
from pathlib import Path
from typing import Any, Dict, Optional, TypeVar

import numpy as np

PathLike = TypeVar("PathLike", str, Path)


def save_metrics(metrics, file_path: PathLike):
    """Save metrics to disk"""
    # convert dtypes to stock Python
    save_metrics = sterilize(metrics)

    # save metrics using safe file write
    file_path_tmp = str(file_path) + '.tmp'
    with open(file_path_tmp, 'w') as fp:
        json.dump(save_metrics, fp, indent=4)
    os.rename(file_path_tmp, file_path)


def load_metrics(file_path: PathLike):
    """Read metrics from disk"""
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"Metrics not found at {file_path}!")

    with open(file_path) as fp:
        metrics = json.load(fp)

    return metrics


def sterilize(obj):
    """Prepare object for conversion to json"""
    if isinstance(obj, dict):
        return {k: sterilize(v) for k, v in obj.items()}
    elif isinstance(obj, (list, tuple, np.ndarray)):
        return [sterilize(v) for v in obj]
    elif isinstance(obj, np.floating):
        return float(obj)
    elif isinstance(obj, np.integer):
        return int(obj)
    elif isinstance(obj, (str, int, bool, float)):
        return obj
    else:
        return obj.__repr__()


def get_dataset_config(
    name: PathLike,
    split: Optional[str] = None,
    fold: Optional[int] = None,
    config_root: PathLike = "/input/dataset-configs",
) -> Dict[str, Any]:
    """
    Read dataset configuration

    Returns:
    - dataset_config: {
        'subject_list': [subject_id1, subject_id2, ...],
        (optional) 'labels': {
            (optional) 'label_name1': {
                subject_id1: 0/1,
                ...
            },
            ...
        }
    }
    """
    config_root = Path(config_root)
    if split is None:
        split = "all"

    if split in ['test', 'all']:
        postfix = ""
    else:
        postfix = f"-fold-{fold}" if fold is not None else ""

    path = config_root / name / f"ds-config-{split}{postfix}.json"
    with open(path) as fp:
        dataset_config = json.load(fp)

    return dataset_config


def get_subject_list(
    name: str,
    split: Optional[str] = None,
    fold: Optional[int] = None,
    config_root: str = "/input/dataset-configs",
    construct_all_from_train=True
) -> Dict[str, Any]:
    """Get subject list from a dataset configuration"""
    try:
        dataset_config = get_dataset_config(name=name, split=split, fold=fold, config_root=config_root)
        return dataset_config['subject_list']
    except FileNotFoundError:
        # if trying to read 'all', construct from train & val splits (assuming 5-fold cross-validation)
        if split == 'all' and construct_all_from_train:
            subject_list_all = []
            for fold in range(5):
                subject_list_all += get_subject_list(name=name, split='train', fold=fold, config_root=config_root)
            subject_list_all = sorted(list(set(subject_list_all)))
            return subject_list_all
        else:
            # could not resolve error, re-raise it
            raise



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pathlib import Path

import numpy as np
import SimpleITK as sitk

try:
    import numpy.typing as npt
except ImportError:  # pragma: no cover
    pass

from picai_eval.data_utils import PathLike


def resize_image_with_crop_or_pad(image, img_size=(64, 64, 64), **kwargs):
    """
    Resize images (scans/predictions/labels) by cropping and/or padding
    Adapted from: https://github.com/DLTK/DLTK]
    """
    assert isinstance(image, np.ndarray)
    assert (image.ndim - 1 == len(img_size) or image.ndim == len(img_size)), \
        "Target size doesn't fit image size"

    rank = len(img_size)  # image dimensions

    # placeholders for new shape
    from_indices = [[0, image.shape[dim]] for dim in range(rank)]
    to_padding = [[0, 0] for _ in range(rank)]
    slicer = [slice(None)] * rank

    # for each dimension, determine process (cropping or padding)
    for i in range(rank):
        if image.shape[i] < img_size[i]:
            to_padding[i][0] = (img_size[i] - image.shape[i]) // 2
            to_padding[i][1] = img_size[i] - image.shape[i] - to_padding[i][0]
        else:
            from_indices[i][0] = int(np.floor((image.shape[i] - img_size[i]) / 2.))
            from_indices[i][1] = from_indices[i][0] + img_size[i]

        # create slicer object to crop/leave each dimension
        slicer[i] = slice(from_indices[i][0], from_indices[i][1])

    # pad cropped image to extend missing dimension
    return np.pad(image[tuple(slicer)], to_padding, **kwargs)


def read_image(path: PathLike):
    """Read image, given a filepath"""
    if isinstance(path, Path):
        path = path.as_posix()
    else:
        assert isinstance(path, str), f"Unexpected path type: {type(path)}. Please provide a Path or str."

    if '.npy' in path:
        return np.load(path)
    elif '.nii' in path or '.mha' in path or 'mhd' in path:
        return sitk.GetArrayFromImage(sitk.ReadImage(path))
    elif '.npz' in path:
        return np.load(path)['softmax'].astype('float32')[1]  # nnUnet format
    else:
        raise ValueError(f"Unexpected file path. Supported file formats: .nii(.gz), .mha, .npy and .npz. Got: {path}.")


def read_prediction(path: PathLike) -> "npt.NDArray[np.float32]":
    """Read prediction, given a filepath"""
    # read prediction and ensure correct dtype
    pred: "npt.NDArray[np.float32]" = np.array(read_image(path), dtype=np.float32)
    return pred


def read_label(path: PathLike) -> "npt.NDArray[np.int32]":
    """Read label, given a filepath"""
    # read label and ensure correct dtype
    lbl: "npt.NDArray[np.int32]" = np.array(read_image(path), dtype=np.int32)
    return lbl



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import concurrent.futures
import itertools
import json
import os
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import (Callable, Dict, Hashable, Iterable, List, Optional, Sized,
                    Tuple, Union)

import numpy as np
from scipy import ndimage
from scipy.optimize import linear_sum_assignment
from tqdm import tqdm

try:
    import numpy.typing as npt
except ImportError:  # pragma: no cover
    pass

from picai_eval.analysis_utils import (calculate_dsc, calculate_iou,
                                       label_structure, parse_detection_map)
from picai_eval.image_utils import (read_label, read_prediction,
                                    resize_image_with_crop_or_pad)
from picai_eval.metrics import Metrics

PathLike = Union[str, Path]


# Compute base prediction metrics TP/FP/FN with associated model confidences
def evaluate_case(
    y_det: "Union[npt.NDArray[np.float32], str, Path]",
    y_true: "Union[npt.NDArray[np.int32], str, Path]",
    min_overlap: float = 0.10,
    overlap_func: "Union[str, Callable[[npt.NDArray[np.float32], npt.NDArray[np.int32]], float]]" = 'IoU',
    case_confidence_func: "Union[str, Callable[[npt.NDArray[np.float32]], float]]" = 'max',
    allow_unmatched_candidates_with_minimal_overlap: bool = True,
    y_det_postprocess_func: "Optional[Callable[[npt.NDArray[np.float32]], npt.NDArray[np.float32]]]" = None,
    y_true_postprocess_func: "Optional[Callable[[npt.NDArray[np.int32]], npt.NDArray[np.int32]]]" = None,
) -> Tuple[List[Tuple[int, float, float]], float]:
    """
    Gather the list of lesion candidates, and classify in TP/FP/FN.

    Lesion candidates are matched to ground truth lesions, by maximizing the number of candidates
    with sufficient overlap (i.e., matches), and secondly by maximizing the total overlap of all candidates.

    Parameters:
    - y_det: Detection map, which should be a 3D volume containing connected components (in 3D) of the
        same confidence. Each detection map may contain an arbitrary number of connected components,
        with different or equal confidences. Alternatively, y_det may be a filename ending in
        .nii.gz/.mha/.mhd/.npy/.npz, which will be loaded on-the-fly.
    - y_true: Ground truth label, which should be a 3D volume of the same shape as the detection map.
        Alternatively, `y_true` may be the filename ending in .nii.gz/.mha/.mhd/.npy/.npz, which should
        contain binary labels and will be loaded on-the-fly. Use `1` to encode ground truth lesion, and
        `0` to encode background.
    - min_overlap: defines the minimal required overlap (e.g., Intersection over Union or Dice similarity
        coefficient) between a lesion candidate and ground truth lesion, to be counted as a true positive
        detection.
    - overlap_func: function to calculate overlap between a lesion candidate and ground truth mask.
        May be 'IoU' for Intersection over Union, or 'DSC' for Dice similarity coefficient. Alternatively,
        provide a function with signature `func(detection_map, annotation) -> overlap [0, 1]`.
    - allow_unmatched_candidates_with_minimal_overlap: when multiple lesion candidates have sufficient
        overlap with the ground truth lesion mask, this determines whether the lesion that is not selected
        counts as a false positive.
    - y_det_postprocess_func: function to apply to detection map. Can for example be used to extract
        lesion candidates from a softmax prediction volume.
    - y_true_postprocess_func: function to apply to annotation. Can for example be used to select the lesion
        masks from annotations that also contain other structures (such as organ segmentations).

    Returns:
    - a list of tuples with:
        (is_lesion, prediction confidence, overlap)
    - case level confidence score derived from the detection map
    """
    y_list: List[Tuple[int, float, float]] = []
    if isinstance(y_true, (str, Path)):
        y_true = read_label(y_true)
    if isinstance(y_det, (str, Path)):
        y_det = read_prediction(y_det)
    if overlap_func == 'IoU':
        overlap_func = calculate_iou
    elif overlap_func == 'DSC':
        overlap_func = calculate_dsc
    elif isinstance(overlap_func, str):
        raise ValueError(f"Overlap function with name {overlap_func} not recognized. Supported are 'IoU' and 'DSC'")

    # convert dtype to float32
    y_true = y_true.astype('int32')
    y_det = y_det.astype('float32')

    # if specified, apply postprocessing functions
    if y_det_postprocess_func is not None:
        y_det = y_det_postprocess_func(y_det)
    if y_true_postprocess_func is not None:
        y_true = y_true_postprocess_func(y_true)

    # check if detection maps need to be padded
    if y_det.shape[0] < y_true.shape[0]:
        print("Warning: padding prediction to match label!")
        y_det = resize_image_with_crop_or_pad(y_det, y_true.shape)
    if np.min(y_det) < 0:
        raise ValueError("All detection confidences must be positive!")

    # perform connected-components analysis on detection maps
    confidences, indexed_pred = parse_detection_map(y_det)
    lesion_candidate_ids = np.arange(len(confidences))

    if not y_true.any():
        # benign case, all predictions are FPs
        for lesion_confidence in confidences.values():
            y_list.append((0, lesion_confidence, 0.))
    else:
        # malignant case, collect overlap between each prediction and ground truth lesion
        labeled_gt, num_gt_lesions = ndimage.label(y_true, structure=label_structure)
        gt_lesion_ids = np.arange(num_gt_lesions)
        overlap_matrix = np.zeros((num_gt_lesions, len(confidences)))

        for lesion_id in gt_lesion_ids:
            # for each lesion in ground-truth (GT) label
            gt_lesion_mask = (labeled_gt == (1+lesion_id))

            # calculate overlap between each lesion candidate and the current GT lesion
            for lesion_candidate_id in lesion_candidate_ids:
                # calculate overlap between lesion candidate and GT mask
                lesion_pred_mask = (indexed_pred == (1+lesion_candidate_id))
                overlap_score = overlap_func(lesion_pred_mask, gt_lesion_mask)

                # store overlap
                overlap_matrix[lesion_id, lesion_candidate_id] = overlap_score

        # match lesion candidates to ground truth lesion (for documentation on how this works, please see
        # https://docs.scipy.org/doc/scipy-0.18.1/reference/generated/scipy.optimize.linear_sum_assignment.html)
        overlap_matrix[overlap_matrix < min_overlap] = 0  # don't match lesions with insufficient overlap
        overlap_matrix[overlap_matrix > 0] += 1  # prioritize matching over the amount of overlap
        matched_lesion_indices, matched_lesion_candidate_indices = linear_sum_assignment(overlap_matrix, maximize=True)

        # remove indices where overlap is zero
        mask = (overlap_matrix[matched_lesion_indices, matched_lesion_candidate_indices] > 0)
        matched_lesion_indices = matched_lesion_indices[mask]
        matched_lesion_candidate_indices = matched_lesion_candidate_indices[mask]

        # all lesion candidates that are matched are TPs
        for lesion_id, lesion_candidate_id in zip(matched_lesion_indices, matched_lesion_candidate_indices):
            lesion_confidence = confidences[lesion_candidate_id]
            overlap = overlap_matrix[lesion_id, lesion_candidate_id]
            overlap -= 1  # return overlap to [0, 1]

            assert overlap > min_overlap, "Overlap must be greater than min_overlap!"

            y_list.append((1, lesion_confidence, overlap))

        # all ground truth lesions that are not matched are FNs
        unmatched_gt_lesions = set(gt_lesion_ids) - set(matched_lesion_indices)
        y_list += [(1, 0., 0.) for _ in unmatched_gt_lesions]

        # all lesion candidates with insufficient overlap/not matched to a gt lesion are FPs
        if allow_unmatched_candidates_with_minimal_overlap:
            candidates_sufficient_overlap = lesion_candidate_ids[(overlap_matrix > 0).any(axis=0)]
            unmatched_candidates = set(lesion_candidate_ids) - set(candidates_sufficient_overlap)
        else:
            unmatched_candidates = set(lesion_candidate_ids) - set(matched_lesion_candidate_indices)
        y_list += [(0, confidences[lesion_candidate_id], 0.) for lesion_candidate_id in unmatched_candidates]

    # determine case-level confidence score
    if case_confidence_func == 'max':
        # take highest lesion confidence as case-level confidence
        case_confidence = np.max(y_det)
    elif case_confidence_func == 'bayesian':
        # if c_i is the probability the i-th lesion is csPCa, then the case-level
        # probability to have one or multiple csPCa lesion is 1 - Π_i{ 1 - c_i}
        case_confidence = 1 - np.prod([(1 - c) for c in confidences.values()])
    else:
        # apply user-defines case-level confidence score function
        case_confidence = case_confidence_func(y_det)

    return y_list, case_confidence


# Evaluate all cases
def evaluate(
    y_det: "Iterable[Union[npt.NDArray[np.float64], str, Path]]",
    y_true: "Iterable[Union[npt.NDArray[np.float64], str, Path]]",
    sample_weight: "Optional[Iterable[float]]" = None,
    subject_list: Optional[Iterable[Hashable]] = None,
    min_overlap: float = 0.10,
    overlap_func: "Union[str, Callable[[npt.NDArray[np.float32], npt.NDArray[np.int32]], float]]" = 'IoU',
    case_confidence_func: "Union[str, Callable[[npt.NDArray[np.float32]], float]]" = 'max',
    allow_unmatched_candidates_with_minimal_overlap: bool = True,
    y_det_postprocess_func: "Optional[Callable[[npt.NDArray[np.float32]], npt.NDArray[np.float32]]]" = None,
    y_true_postprocess_func: "Optional[Callable[[npt.NDArray[np.int32]], npt.NDArray[np.int32]]]" = None,
    num_parallel_calls: int = 3,
    verbose: int = 0,
) -> Metrics:
    """
    Evaluate 3D detection performance.

    Parameters:
    - y_det: iterable of all detection_map volumes to evaluate. Each detection map should a 3D volume
        containing connected components (in 3D) of the same confidence. Each detection map may contain
        an arbitrary number of connected components, with different or equal confidences.
        Alternatively, y_det may contain filenames ending in .nii.gz/.mha/.mhd/.npy/.npz, which will
        be loaded on-the-fly.
    - y_true: iterable of all ground truth labels. Each label should be a 3D volume of the same shape
        as the corresponding detection map. Alternatively, `y_true` may contain filenames ending in
        .nii.gz/.mha/.mhd/.npy/.npz, which should contain binary labels and will be loaded on-the-fly.
        Use `1` to encode ground truth lesion, and `0` to encode background.
    - sample_weight: case-level sample weight. These weights will also be applied to the lesion-level
        evaluation, with same weight for all lesion candidates of the same case.
    - subject_list: list of sample identifiers, to give recognizable names to the evaluation results.
    - min_overlap: defines the minimal required Intersection over Union (IoU) or Dice similarity
        coefficient (DSC) between a lesion candidate and ground truth lesion, to be counted as a true
        positive detection.
    - overlap_func: function to calculate overlap between a lesion candidate and ground truth mask.
        May be 'IoU' for Intersection over Union, or 'DSC' for Dice similarity coefficient. Alternatively,
        provide a function with signature `func(detection_map, annotation) -> overlap [0, 1]`.
    - case_confidence_func: function to derive case-level confidence from detection map. Default: max.
    - allow_unmatched_candidates_with_minimal_overlap: when multiple lesion candidates have sufficient
        overlap with the ground truth lesion mask, this determines whether the lesion that is not selected
        counts as a false positive.
    - y_det_postprocess_func: function to apply to detection map. Can for example be used to extract
        lesion candidates from a softmax prediction volume.
    - y_true_postprocess_func: function to apply to annotation. Can for example be used to select the lesion
        masks from annotations that also contain other structures (such as organ segmentations).
    - num_parallel_calls: number of threads to use for evaluation.
    - verbose: (optional) controll amount of printed information.

    Returns:
    - Metrics
    """
    if sample_weight is None:
        sample_weight = itertools.repeat(1)
    if subject_list is None:
        # generate indices to keep track of each case during multiprocessing
        subject_list = itertools.count()

    # initialize placeholders
    case_target: Dict[Hashable, int] = {}
    case_weight: Dict[Hashable, float] = {}
    case_pred: Dict[Hashable, float] = {}
    lesion_results: Dict[Hashable, List[Tuple[int, float, float]]] = {}
    lesion_weight: Dict[Hashable, List[float]] = {}

    with ThreadPoolExecutor(max_workers=num_parallel_calls) as pool:
        # define the functions that need to be processed: compute_pred_vector, with each individual
        # detection_map prediction, ground truth label and parameters
        future_to_args = {
            pool.submit(
                evaluate_case,
                y_det=y_det_case,
                y_true=y_true_case,
                min_overlap=min_overlap,
                overlap_func=overlap_func,
                case_confidence_func=case_confidence_func,
                allow_unmatched_candidates_with_minimal_overlap=allow_unmatched_candidates_with_minimal_overlap,
                y_det_postprocess_func=y_det_postprocess_func,
                y_true_postprocess_func=y_true_postprocess_func
            ): (idx, weight)
            for (y_det_case, y_true_case, weight, idx) in zip(y_det, y_true, sample_weight, subject_list)
        }

        # process the cases in parallel
        iterator = concurrent.futures.as_completed(future_to_args)
        if verbose:
            total: Optional[int] = None
            if isinstance(subject_list, Sized):
                total = len(subject_list)
            iterator = tqdm(iterator, desc='Evaluating', total=total)

        for future in iterator:
            idx, weight = future_to_args[future]

            try:
                # unpack results
                lesion_results_case, case_confidence = future.result()
            except Exception as e:
                print(f'Error for {idx}: {e}')
                raise e

            # aggregate results
            idx, weight = future_to_args[future]
            case_weight[idx] = weight
            case_pred[idx] = case_confidence
            if len(lesion_results_case):
                case_target[idx] = np.max([a[0] for a in lesion_results_case])
            else:
                case_target[idx] = 0

            # accumulate outputs
            lesion_results[idx] = lesion_results_case
            lesion_weight[idx] = [weight] * len(lesion_results_case)

    # collect results in a Metrics object
    metrics = Metrics(
        lesion_results=lesion_results,
        case_target=case_target,
        case_pred=case_pred,
        case_weight=case_weight,
        lesion_weight=lesion_weight,
    )

    return metrics


def evaluate_folder(
    y_det_dir: Union[Path, str],
    y_true_dir: Optional[Union[Path, str]] = None,
    subject_list: Optional[Union[List[str], PathLike]] = None,
    pred_extensions: Optional[List[str]] = None,
    label_extensions: Optional[List[str]] = None,
    detection_map_postfixes: Optional[List[str]] = None,
    label_postfixes: Optional[List[str]] = None,
    verbose: int = 1,
    **kwargs
) -> Metrics:
    """
    Evaluate 3D detection performance, for all samples in y_det_dir,
    or the samples specified in the subject_list.

    Parameters:
    - y_det_dir: path to folder containing the detection maps.
    - y_true_dir: (optional) path to folder containing the annotations. Defaults to y_true_dir.
    - subject_list: (optional) list of cases to evaluate. Allows to evaluate a subset of cases in a folder,
                    and ensures all specified cases were found. If str or Path, will load the subject list
                    from the provided JSON file, which should contain a dictionary with "subject_list" entry.
    - pred_extensions: allowed file extensions for detection maps.
    - label_extensions: allowed file extensions for annotations.
    - detection_map_postfixes: allowed postifxes for detection maps.
    - label_postfixes: allowed postifxes for annotations.
    - verbose: (optional) controll amount of printed information.
    **kwargs: (optional) see `evaluate` for additional options

    Returns:
    - Metrics
    """
    if y_true_dir is None:
        y_true_dir = y_det_dir
    if pred_extensions is None:
        pred_extensions = [".npz", ".npy", ".nii.gz", ".nii", ".mha", ".mhd"]
    if label_extensions is None:
        label_extensions = [".nii.gz", ".nii", ".mha", ".mhd", ".npz", ".npy"]
    if detection_map_postfixes is None:
        detection_map_postfixes = ["_detection_map"]
        if y_true_dir != y_det_dir:
            # if annotation directory is specified, also look for [subject_id].nii.gz etc
            detection_map_postfixes += [""]
    if label_postfixes is None:
        label_postfixes = ["_label"]
        if y_true_dir != y_det_dir:
            # if annotation directory is specified, also look for [subject_id].nii.gz etc
            label_postfixes += [""]
    if isinstance(subject_list, (str, Path)):
        with open(subject_list) as fp:
            subject_list = json.load(fp)["subject_list"]

    # combine postfixes and extensions in a single list
    detection_map_postfixes = [
        f"{postfix}{extension}"
        for postfix in detection_map_postfixes
        for extension in pred_extensions
    ]
    label_postfixes = [
        f"{postfix}{extension}"
        for postfix in label_postfixes
        for extension in label_extensions
    ]

    y_det = []
    y_true = []
    if subject_list:
        # collect the detection maps and labels for each case specified in the subject list
        for subject_id in subject_list:
            # construct paths to detection maps and labels
            found_pred, found_label = False, False
            for postfix in detection_map_postfixes:
                detection_path = os.path.join(y_det_dir, f"{subject_id}{postfix}")
                if os.path.exists(detection_path):
                    y_det += [detection_path]
                    found_pred = True
                    break

            for postfix in label_postfixes:
                label_path = os.path.join(y_true_dir, f"{subject_id}{postfix}")
                if os.path.exists(label_path):
                    y_true += [label_path]
                    found_label = True
                    break

            # check if detection map and label are found
            assert found_pred, f"Did not find prediction for {subject_id} in {y_det_dir}!"
            assert found_label, f"Did not find label for {subject_id} in {y_true_dir}!"
    else:
        # collect all detection maps found in detection_map_dir
        file_list = sorted(os.listdir(y_det_dir))
        subject_list = []
        if verbose >= 1:
            print(f"Found {len(file_list)} files in the input directory, collecting detection_mapes with " +
                  f"{detection_map_postfixes} and labels with {label_postfixes}.")

        # collect filenames of detection_map predictions
        for fn in file_list:
            for postfix in detection_map_postfixes:
                if postfix in fn:
                    y_det += [os.path.join(y_det_dir, fn)]
                    subject_id = fn.replace(postfix, "")
                    if subject_id in subject_list:
                        print(f"Found multiple detection maps for {subject_id}, skipping {fn}!")
                        continue

                    subject_list += [subject_id]
                    break

        # collect filenames of labels
        for subject_id in subject_list:
            found_label = False
            for postfix in label_postfixes:
                label_path = os.path.join(y_true_dir, f"{subject_id}{postfix}")
                if os.path.exists(label_path):
                    y_true += [label_path]
                    found_label = True
                    break

            assert found_label, f"Did not find label for {subject_id} in {y_true_dir}!"

    # ensure files exist
    for subject_id, detection_path in zip(subject_list, y_det):
        assert os.path.exists(detection_path), f"Could not find prediction for {subject_id} at {detection_path}!"
    for subject_id, label_path in zip(subject_list, y_true):
        assert os.path.exists(label_path), f"Could not find label for {subject_id} at {label_path}!"

    if verbose >= 1:
        print(f"Found prediction and label for {len(y_det)} cases. Here are some examples:")
        print(subject_list[0:5])

    # check if predictions were found
    assert len(y_det), f"Did not find any predictions in {y_det_dir}!"

    # perform evaluation with compiled file lists
    return evaluate(y_det=y_det, y_true=y_true, subject_list=subject_list, verbose=verbose, **kwargs)



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import argparse
import json
import os

from picai_eval import evaluate_folder

# acquire and parse input and output paths
parser = argparse.ArgumentParser(description='Command Line Arguments')
parser.add_argument("-i", "--input", type=str, required=True,
                    help="Path to folder with model predicitons (detection maps)")
parser.add_argument("-l", "--labels", type=str, required=False,
                    help="Path to folder with labels (defaults to input folder if unspecified)")
parser.add_argument("-o", "--output", type=str, default="metrics.json",
                    help="Path to store metrics file, relative to the input folder.")
parser.add_argument("-s", "--subject_list", type=str, required=False,
                    help="Path to subject list, relative to the input folder. The subject list " +
                         "may be stored as json list, or json dictionary with 'subject_list' as parameter.")
parser.add_argument("--pred_extensions", type=str, nargs="+", required=False,
                    help="List of allowed file formats for detection maps." +
                         "Default: .npz, .npy, .nii.gz, .nii, .mha and .mhd")
parser.add_argument("--label_extensions", type=str, nargs="+", required=False,
                    help="List of allowed file formats for annotations." +
                         "Default: .nii.gz, .nii, .mha, .mhd, .npz and .npy")
args = parser.parse_args()

if args.labels is None:
    args.labels = args.input
args.output = os.path.join(args.input, args.output)
if args.subject_list is not None:
    args.subject_list = os.path.join(args.input, args.subject_list)
    with open(args.subject_list) as fp:
        args.subject_list = json.load(fp)
    if isinstance(args.subject_list, dict):
        args.subject_list = args.subject_list['subject_list']

print(f"""
    PICAI Evaluation
    Model predictions path: {args.input}
    Labels path: {args.labels}
    Output Metrics Path: {args.output}
""")

assert os.path.exists(args.input), f"Input folder does not exist at {args.input}!"
assert os.path.exists(os.path.dirname(args.output)), f"Output folder does not exist at {args.output}!"

# calculate metrics
metrics = evaluate_folder(
    y_det_dir=args.input,
    y_true_dir=args.labels,
    subject_list=args.subject_list,
    pred_extensions=args.pred_extensions,
    label_extensions=args.label_extensions,
)

# show metrics
print("Metrics:")
print(metrics)

# save metrics
metrics.save(path=args.output)



import os
from pathlib import Path
from subprocess import check_call

import setuptools
from setuptools.command.develop import develop
from setuptools.command.install import install
from setuptools.command.sdist import sdist

HERE = Path(__file__).parent


def gitcmd_update_submodules():
    """
    Check if the package is being deployed as a git repository. If so, recursively
    update all dependencies.

        Returns:
        True if the package is a git repository and the modules were updated. False otherwise.
    """
    if os.path.exists(os.path.join(HERE, '.git')):
        check_call(['git', 'submodule', 'update', '--init', '--recursive'])
        return True

    return False


class gitcmd_develop(develop):
    """
    Specialized packaging class that runs git submodule update --init --recursive
    as part of the update/install procedure.
    """

    def run(self):
        gitcmd_update_submodules()
        develop.run(self)


class gitcmd_install(install):
    """
    Specialized packaging class that runs git submodule update --init --recursive
    as part of the update/install procedure.
    """

    def run(self):
        gitcmd_update_submodules()
        install.run(self)


class gitcmd_sdist(sdist):
    """
    Specialized packaging class that runs git submodule update --init --recursive
    as part of the update/install procedure;.
    """

    def run(self):
        gitcmd_update_submodules()
        sdist.run(self)


if __name__ == '__main__':
    with open("README.md", "r", encoding="utf-8") as fh:
        long_description = fh.read()

    setuptools.setup(
        version='1.4.3',  # also update version in metrics.py -> version
        author_email='Joeran.Bosma@radboudumc.nl',
        long_description=long_description,
        long_description_content_type="text/markdown",
        url='https://github.com/DIAGNijmegen/picai_eval',
        project_urls={
            "Bug Tracker": "https://github.com/DIAGNijmegen/picai_eval/issues"
        },
        license='Apache 2.0',
        packages=['picai_eval', 'picai_eval.stat_util'],
        cmdclass={
            'develop': gitcmd_develop,
            'install': gitcmd_install,
            'sdist': gitcmd_sdist,
        },
    )
